/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.ws.jws2;

import jalview.bin.Cache;
import jalview.bin.ApplicationSingletonProvider;
import jalview.bin.ApplicationSingletonProvider.ApplicationSingletonI;
import jalview.gui.AlignFrame;
import jalview.util.MessageManager;
import jalview.ws.ServiceChangeListener;
import jalview.ws.WSDiscovererI;
import jalview.ws.api.ServiceWithParameters;
import jalview.ws.jws2.jabaws2.Jws2Instance;
import jalview.ws.params.ParamDatastoreI;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.CopyOnWriteArraySet;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.FutureTask;

import javax.swing.JMenu;

import compbio.ws.client.Services;

/**
 * discoverer for jws2 services. Follows the lightweight service discoverer
 * pattern (archetyped by EnfinEnvision2OneWay)
 * 
 * @author JimP
 * 
 */
public class Jws2Discoverer implements WSDiscovererI, Runnable, ApplicationSingletonI
{

  /**
   * Returns the singleton instance of this class.
   * 
   * @return
   */
  public static Jws2Discoverer getInstance()
  {
    return (Jws2Discoverer) ApplicationSingletonProvider
            .getInstance(Jws2Discoverer.class);
  }

  public static final String COMPBIO_JABAWS = "http://www.compbio.dundee.ac.uk/jabaws";

  /*
   * the .jalview_properties entry for JWS2 URLS
   */
  private final static String JWS2HOSTURLS = "JWS2HOSTURLS";

  /*
   * Override for testing only
   */
  private List<String> testUrls = null;

  // preferred url has precedence over others
  private String preferredUrl;
  
  private Set<ServiceChangeListener> serviceListeners = new CopyOnWriteArraySet<>();
  private Vector<String> invalidServiceUrls = null;

  private Vector<String> urlsWithoutServices = null;

  private Vector<String> validServiceUrls = null;

  private volatile boolean running = false;

  private volatile boolean aborted = false;
  
  private volatile Thread oldthread = null;

  /**
   * holds list of services.
   */
  protected Vector<Jws2Instance> services;

  /**
   * Private constructor enforces use of singleton via getDiscoverer()
   */
  private Jws2Discoverer()
  {
  }


  @Override
  public void addServiceChangeListener(ServiceChangeListener listener)
  {
    serviceListeners.add(listener);
  }

  @Override
  public void removeServiceChangeListener(ServiceChangeListener listener)
  {
    serviceListeners.remove(listener);
  }

  private void notifyServiceListeners(List<? extends ServiceWithParameters> services) 
  {
    if (services == null) services = this.services;
    for (var listener : serviceListeners) {
      listener.servicesChanged(this, services);
    }
  }

  /**
   * @return the aborted
   */
  public boolean isAborted()
  {
    return aborted;
  }

  /**
   * @param aborted
   *          the aborted to set
   */
  public void setAborted(boolean aborted)
  {
    this.aborted = aborted;
  }

  @Override
  public void run()
  {

    if (running && oldthread != null && oldthread.isAlive())
    {
      if (!aborted)
      {
        return;
      }
      while (running)
      {
        try
        {
          Cache.log.debug(
                  "Waiting around for old discovery thread to finish.");
          // wait around until old discoverer dies
          Thread.sleep(100);
        } catch (Exception e)
        {
        }
      }
      aborted = false;
      Cache.log.debug("Old discovery thread has finished.");
    }
    running = true;

    // first set up exclusion list if needed
    final Set<String> ignoredServices = new HashSet<>();
    for (String ignored : Cache
            .getDefault("IGNORED_JABAWS_SERVICETYPES", "").split("\\|"))
    {
      ignoredServices.add(ignored);
    }

    notifyServiceListeners(Collections.emptyList());
    oldthread = Thread.currentThread();
    try
    {
      getClass().getClassLoader().loadClass("compbio.ws.client.Jws2Client");
    } catch (ClassNotFoundException e)
    {
      System.err.println(
              "Not enabling JABA Webservices : client jar is not available."
                      + "\nPlease check that your webstart JNLP file is up to date!");
      running = false;
      return;
    }
    // reinitialise records of good and bad service URLs
    if (services != null)
    {
      services.removeAllElements();
    }
    if (urlsWithoutServices != null)
    {
      urlsWithoutServices.removeAllElements();
    }
    if (invalidServiceUrls != null)
    {
      invalidServiceUrls.removeAllElements();
    }
    if (validServiceUrls != null)
    {
      validServiceUrls.removeAllElements();
    }
    ArrayList<String> svctypes = new ArrayList<>();

    List<JabaWsServerQuery> qrys = new ArrayList<>();
    for (final String jwsserver : getServiceUrls())
    {
      JabaWsServerQuery squery = new JabaWsServerQuery(this, jwsserver);
      if (svctypes.size() == 0)
      {
        // TODO: remove this ugly hack to get Canonical JABA service ordering
        // for all possible services
        for (Services sv : squery.JABAWS2SERVERS)
        {
          if (!ignoredServices.contains(sv.toString()))
          {
            svctypes.add(sv.toString());
          }
        }

      }
      qrys.add(squery);
      new Thread(squery).start();
    }
    boolean finished = true;
    do
    {
      finished = true;
      try
      {
        Thread.sleep(100);
      } catch (Exception e)
      {
      }
      for (JabaWsServerQuery squery : qrys)
      {
        if (squery.isRunning())
        {
          finished = false;
        }
      }
      if (aborted)
      {
        Cache.log.debug(
                "Aborting " + qrys.size() + " JABAWS discovery threads.");
        for (JabaWsServerQuery squery : qrys)
        {
          squery.setQuit(true);
        }
      }
    } while (!aborted && !finished);
    if (!aborted)
    {
      // resort services according to order found in jabaws service list
      // also ensure services for each host are ordered in same way.

      if (services != null && services.size() > 0)
      {
        Jws2Instance[] svcs = new Jws2Instance[services.size()];
        int[] spos = new int[services.size()];
        int ipos = 0;
        List<String> svcUrls = getServiceUrls();
        for (Jws2Instance svc : services)
        {
          svcs[ipos] = svc;
          spos[ipos++] = 1000 * svcUrls.indexOf(svc.getHostURL()) + 1
                  + svctypes.indexOf(svc.getName());
        }
        jalview.util.QuickSort.sort(spos, svcs);
        services = new Vector<>();
        for (Jws2Instance svc : svcs)
        {
          if (!ignoredServices.contains(svc.getName()))
          {
            services.add(svc);
          }
        }
      }
    }
    oldthread = null;
    running = false;
    notifyServiceListeners(services);
  }

  /**
   * record this service endpoint so we can use it
   * 
   * @param jwsservers
   * @param srv
   * @param service2
   */
  synchronized void addService(String jwsservers, Jws2Instance service)
  {
    if (services == null)
    {
      services = new Vector<>();
    }
    System.out.println(
            "Discovered service: " + jwsservers + " " + service.toString());
    // Jws2Instance service = new Jws2Instance(jwsservers, srv.toString(),
    // service2);

    services.add(service);
    // retrieve the presets and parameter set and cache now
    ParamDatastoreI pds = service.getParamStore();
    if (pds != null)
    {
      pds.getPresets();
    }
    service.hasParameters();
    if (validServiceUrls == null)
    {
      validServiceUrls = new Vector<>();
    }
    validServiceUrls.add(jwsservers);
  }

  /**
   * 
   * @param args
   * @j2sIgnore
   */
  public static void main(String[] args)
  {
    Jws2Discoverer instance = getInstance();
    if (args.length > 0)
    {
      instance.testUrls = new ArrayList<>();
      for (String url : args)
      {
        instance.testUrls.add(url);
      }
    }
    var discoverer = getInstance();
    discoverer.addServiceChangeListener((_discoverer, _services) -> {
      if (discoverer.services != null)
      {
        System.out.println("Changesupport: There are now "
                + discoverer.services.size() + " services");
        int i = 1;
        for (ServiceWithParameters s_instance : discoverer.services)
        {
          System.out.println(
                  "Service " + i++ + " " + s_instance.getClass()
                          + "@" + s_instance.getHostURL() + ": "
                          + s_instance.getActionText());
        }

      }
    });
    try
    {
      discoverer.startDiscoverer().get();
    } catch (InterruptedException | ExecutionException e)
    {
    }
    try
    {
      Thread.sleep(50);
    } catch (InterruptedException x)
    {
    }
  }


  @Override
  public boolean hasServices()
  {
    return !running && services != null && services.size() > 0;
  }

  @Override
  public boolean isRunning()
  {
    return running;
  }

  @Override
  public void setServiceUrls(List<String> wsUrls)
  {
    if (wsUrls != null && !wsUrls.isEmpty())
    {
      StringBuilder urls = new StringBuilder(128);
      String sep = "";
      for (String url : wsUrls)
      {
        urls.append(sep);
        urls.append(url);
        sep = ",";
      }
      Cache.setProperty(JWS2HOSTURLS, urls.toString());
    }
    else
    {
      Cache.removeProperty(JWS2HOSTURLS);
    }
  }

  /**
   * Returns web service URLs, in the order in which they should be tried (or an
   * empty list).
   * 
   * @return
   */
  @Override
  public List<String> getServiceUrls()
  {
    if (testUrls != null)
    {
      // return test urls, if there are any, instead of touching cache
      return testUrls;
    }
    List<String> urls = new ArrayList<>();

    if (this.preferredUrl != null)
    {
      urls.add(preferredUrl);
    }

    String surls = Cache.getDefault(JWS2HOSTURLS, COMPBIO_JABAWS);
    try
    {
      StringTokenizer st = new StringTokenizer(surls, ",");
      while (st.hasMoreElements())
      {
        String url = null;
        try
        {
          url = st.nextToken();
          new URL(url);
          if (!urls.contains(url))
          {
            urls.add(url);
          }
          else
          {
            Cache.log.warn("Ignoring duplicate url " + url + " in "
                    + JWS2HOSTURLS + " list");
          }
        } catch (MalformedURLException ex)
        {
          Cache.log.warn("Problem whilst trying to make a URL from '"
                  + ((url != null) ? url : "<null>") + "'");
          Cache.log.warn(
                  "This was probably due to a malformed comma separated list"
                          + " in the " + JWS2HOSTURLS
                          + " entry of $(HOME)/.jalview_properties)");
          Cache.log.debug("Exception was ", ex);
        }
      }
    } catch (Exception ex)
    {
      Cache.log.warn("Error parsing comma separated list of urls in "
              + JWS2HOSTURLS + " preference.", ex);
    }
    return urls;
  }

  @Override
  public Vector<ServiceWithParameters> getServices()
  {
    return (services == null) ? new Vector<>() : new Vector<>(services);
  }

  /**
   * test the given URL with the JabaWS test code
   * 
   * @param foo
   * @return
   */
  @Override
  public boolean testServiceUrl(URL foo)
  {
    try
    {
      compbio.ws.client.WSTester
              .main(new String[]
              { "-h=" + foo.toString() });
    } catch (Exception e)
    {
      e.printStackTrace();
      return false;
    } catch (OutOfMemoryError e)
    {
      e.printStackTrace();
      return false;
    } catch (Error e)
    {
      e.printStackTrace();
      return false;
    }

    return true;
  }

  public boolean restart()
  {
    synchronized (this)
    {
      if (running)
      {
        aborted = true;
      }
      else
      {
        running = true;
      }
      return aborted;
    }
  }

  /**
   * Start a fresh discovery thread and notify the given object when we're
   * finished. Any known existing threads will be killed before this one is
   * started.
   * 
   * @param changeSupport2
   * @return new thread
   */
  @Override
  public CompletableFuture<WSDiscovererI> startDiscoverer()
  {
    /*    if (restart())
        {
          return;
        }
        else
        {
          Thread thr = new Thread(this);
          thr.start();
        }
       */
    if (isRunning())
    {
      setAborted(true);
    }
    CompletableFuture<WSDiscovererI> task = CompletableFuture
            .supplyAsync(() -> {
              run();
              return Jws2Discoverer.this;
            });
    return task;
  }

  /**
   * @return the invalidServiceUrls
   */
  public Vector<String> getInvalidServiceUrls()
  {
    return invalidServiceUrls;
  }

  /**
   * @return the urlsWithoutServices
   */
  public Vector<String> getUrlsWithoutServices()
  {
    return urlsWithoutServices;
  }

  /**
   * add an 'empty' JABA server to the list. Only servers not already in the
   * 'bad URL' list will be added to this list.
   * 
   * @param jwsservers
   */
  public synchronized void addUrlwithnoservices(String jwsservers)
  {
    if (urlsWithoutServices == null)
    {
      urlsWithoutServices = new Vector<>();
    }

    if ((invalidServiceUrls == null
            || !invalidServiceUrls.contains(jwsservers))
            && !urlsWithoutServices.contains(jwsservers))
    {
      urlsWithoutServices.add(jwsservers);
    }
  }

  /**
   * add a bad URL to the list
   * 
   * @param jwsservers
   */
  public synchronized void addInvalidServiceUrl(String jwsservers)
  {
    if (invalidServiceUrls == null)
    {
      invalidServiceUrls = new Vector<>();
    }
    if (!invalidServiceUrls.contains(jwsservers))
    {
      invalidServiceUrls.add(jwsservers);
    }
  }

  /**
   * 
   * @return a human readable report of any problems with the service URLs used
   *         for discovery
   */
  @Override
  public String getErrorMessages()
  {
    if (!isRunning() && !isAborted())
    {
      StringBuffer ermsg = new StringBuffer();
      boolean list = false;
      if (getInvalidServiceUrls() != null
              && getInvalidServiceUrls().size() > 0)
      {
        ermsg.append(MessageManager.getString("warn.urls_not_contacted")
                + ": \n");
        for (String svcurl : getInvalidServiceUrls())
        {
          if (list)
          {
            ermsg.append(", ");
          }
          list = true;
          ermsg.append(svcurl);
        }
        ermsg.append("\n\n");
      }
      list = false;
      if (getUrlsWithoutServices() != null
              && getUrlsWithoutServices().size() > 0)
      {
        ermsg.append(
                MessageManager.getString("warn.urls_no_jaba") + ": \n");
        for (String svcurl : getUrlsWithoutServices())
        {
          if (list)
          {
            ermsg.append(", ");
          }
          list = true;
          ermsg.append(svcurl);
        }
        ermsg.append("\n");
      }
      if (ermsg.length() > 1)
      {
        return ermsg.toString();
      }

    }
    return null;
  }

  @Override
  public int getServerStatusFor(String url)
  {
    if (validServiceUrls != null && validServiceUrls.contains(url))
    {
      return STATUS_OK;
    }
    if (urlsWithoutServices != null && urlsWithoutServices.contains(url))
    {
      return STATUS_NO_SERVICES;
    }
    if (invalidServiceUrls != null && invalidServiceUrls.contains(url))
    {
      return STATUS_INVALID;
    }
    return STATUS_UNKNOWN;
  }

  /**
   * Set a URL to try before any others. For use with command-line parameter to
   * configure a local Jabaws installation without the need to add to property
   * files.
   * 
   * @param value
   * @throws MalformedURLException
   */
  public void setPreferredUrl(String value) throws MalformedURLException
  {
    if (value != null && value.trim().length() > 0)
    {
      new URL(value);
      preferredUrl = value;
    }
  }
}
